#!/usr/bin/perl

$|=1;

use strict;
use Time::Local;
use Getopt::Long ;
Getopt::Long::Configure('ignore_case') ;

# to be supplied via args:
my ($narrdir,$rundir,@domains,$inittime,$interval,$fcstlen,$verbose,$debug,$tmpdir,$NWPDIR,$wrfpostroot,
   @types,$dowrfpost,$donarr,$donative,$dostageiv,$stageivdir);

$narrdir='/data/ncep-data/narr';
$stageivdir='/anvil/carbon/precip-stageiv/data/';
$interval=10800;
$fcstlen=30;
$verbose=1;
$debug=0;
$tmpdir=`pwd`;
chomp $tmpdir;
$NWPDIR="$ENV{HOME}/NWP";
$wrfpostroot="/squall/dap02apr180gb/p1229-carbon/wrf/wrfpost2.0";
$dowrfpost=1;
$donarr=1;
$dostageiv=1;
$donative=1;

my %options = ('verbose!' => \$verbose,       
	       'debug!' => \$debug,       
	       'interval=i' => \$interval,
	       'inittime=i' => \$inittime,
	       'fcstlen=i' => \$fcstlen,
	       'domains=i@' => \@domains,
	       'rundir=s' => \$rundir,
	       'narrdir=s' => \$narrdir,
	       'stageivdir=s' => \$stageivdir,
	       'NWPDIR=s' => \$NWPDIR,
	       'tmpdir=s' => \$tmpdir,
	       'wrfpostroot=s' => \$wrfpostroot,
	       'types=s@' => \@types,
	       'dowrfpost!' => \$dowrfpost,
	       'donarr!' => \$donarr,
	       'dostageiv!' => \$dostageiv,
	       'donative!' => \$donative);
my %helpopt = ('verbose!' => "turn on diagnostic output",
	       'debug!'=> "turn on debugging",
	       'interval=is' => "output interval (sec)",
	       'inittime=i' => "REQ: initial time of forecast (yyyymmddhh)",
	       'fcstlen=i' => "length of forecast (hours)",
	       'domains=i@' => "Domain numbers to process",
	       'rundir=s' => "REQ: Archive directory for WRF grib files, e.g. /project/p1229/wrf/p1229-case06/wrfout",
	       'narrdir=s' => "Directory containing narr files",
	       'stageivdir=s' => "Directory containing stageiv files",
	       'NWPDIR=s' => "parent directory of native cvs working directories",
	       'tmpdir=s' => "Working directory for this run",
	       'wrfpostroot=s' => "directory containing run_wrfpost.pl and wrfpost codes (REQ if dowrfpost)",
	       'types=s@' => "types of native output files to produce",
	       'dowrfpost!' => "run wrfpost",
	       'donarr!' => "process narr files",
	       'dostageiv!' => "process stageiv files",
	       'donative!' => "process wrfpost output files");

#my $help=1 unless GetOptions (%options);
my $help;
GetOptions (%options);
$help = 1 if (!$inittime);
$help = 1 if (!$rundir &&  $dowrfpost);
$help = 1 if (!$wrfpostroot && $dowrfpost);
$verbose = 1 if ($debug || $help);
@types = ('all_100_all_3d','uv_105_10_2d','all_105_2_norot','all_1_all_norot','pcp_1_all_norot') unless @types;
@domains=(1,2) unless @domains;

my @call = caller(0);
if (@call > 6) {
    $0 = $call[6] if ($call[3] eq "(eval)");
}
if ($verbose) {
  print "At ",scalar(localtime)," running $0 with:\n";
  foreach my $key (keys(%options)) {
    if (substr($key,length($key)-1) eq "@") {
      print "  $key = @{$options{$key}}\n";
    } else {
      print "  $key = ${$options{$key}}\n";
    }
  }
}

&help(\%options, \%helpopt) if $help;
chdir $tmpdir or die "Cannot chdir $tmpdir\n";

my @typ_2d3d=();
my @typ_ind=();
for (my $i=0;$i<@types;$i++) {
  $typ_2d3d[$i]='norot';
  $typ_ind[$i]=$i;
}
&mkctl(\@types,\@typ_2d3d,\@typ_ind) if ($donarr || $donative);

if ($donarr) {
  mkdir ("narr") or warn ("Could not mkdir $tmpdir/narr\n");
}
if ($dostageiv) {
  mkdir ("stageiv") or warn ("Could not mkdir $tmpdir/stageiv\n");
}

if ($dowrfpost || $donative) {
  mkdir ("$inittime") or warn ("Could not mkdir $tmpdir/$inittime\n");
}

my $cmd;
my $verbose_str = $verbose ? "-verbose" : "-noverbose";
my $debug_str = $debug ? "-debug" : "-nodebug";
if ($dowrfpost) {
  chdir ("$inittime") or die ("Could chdir $tmpdir/$inittime\n");
  $cmd="$wrfpostroot/run_wrfpost.pl -indir $rundir/$inittime -inittime $inittime -intervals $interval";
  $cmd="$cmd -wrfpostroot $wrfpostroot";
  foreach my $domain (@domains) {
    $cmd="$cmd -domains $domain";
  }
  $cmd="$cmd $verbose_str $debug_str";
  print "At ",scalar localtime()," Running $cmd\n" if $verbose;
  die ("Error running $cmd\n") if (system($cmd));
}

my ($iyr,$imo,$ida,$ihr,$imin,$isec);
if ($inittime =~ /(\d\d\d\d)(\d\d)(\d\d)(\d\d)/ ) {
  $iyr=$1;
  $imo=$2;
  $ida=$3;
  $ihr=$4;
  $imin=0;
  $isec=0;
} else {
  die("Invalid inittime=$inittime\n");
}
my $initepoch=timegm($isec,$imin,$ihr,$ida,$imo-1,$iyr-1900);
print "inittime=$inittime (Converted to epoch and back:".scalar(gmtime($initepoch)).")\n" if $verbose;

my $itime=0;
my $leadsec=0;
my $fcstsec=3600*$fcstlen;

# Loop over all forecast times for this forecast
while ( $leadsec <= $fcstsec ) {
  my $hour=int($leadsec/3600);
  my $minute=int(($leadsec-3600*$hour)/60+0.5);
  my $shour=sprintf("%02d",$hour);
  my $smin=sprintf("%02d",$minute);
  my $vepoch=$initepoch + $leadsec;
  my ($vsec,$vmin,$vhour,$vmday,$vmon,$vyear,$vwday,$vyday) = gmtime($vepoch);
  my $vsyear=1900+$vyear;
  my $vsmon=sprintf("%02d",$vmon+1);
  my $vsday=sprintf("%02d",$vmday);
  my $vshour=sprintf("%02d",$vhour);
  my $vsmin=sprintf("%02d",$vmin);
  my $vssec=sprintf("%02d",$vsec);
  if ($donarr) {
    chdir "$tmpdir/narr" or die ("Could chdir $tmpdir/narr\n");
    my $narrin="$narrdir/narr-a_221_${vsyear}${vsmon}${vsday}_${vshour}${vsmin}_000.grb";
    #Loop over types of native files:
    my $reuse=" ";
    foreach my $ctl (@types) {
      my $narrout="narr-a_221_${vsyear}${vsmon}${vsday}_${vshour}${vsmin}_${ctl}";
      if (! -e "${narrout}.hdr" || ! -e "${narrout}.dat") {
	$cmd="$NWPDIR/native/grib_to_native.pl -gribfile $narrin -ctlnam ../${ctl}.in -wgribopt ieee -nativefile $narrout";
	$cmd="$cmd $verbose_str $debug_str $reuse";
	print "At ",scalar localtime()," Running $cmd\n" if $verbose;
	warn ("Error running $cmd\n") if (system($cmd));
	$reuse="-reuse_gribi";
      }
    }
  }
  if ($dostageiv && ($vhour % 6) == 0) {
    chdir "$tmpdir/stageiv" or die ("Could chdir $tmpdir/stageiv\n");
    my $stageivin="$stageivdir/ST4.${vsyear}${vsmon}${vsday}${vshour}.06h";
    #Loop over types of native files:
    my $ctl='pcp_1_all_norot';
    my $stageivout="ST4.${vsyear}${vsmon}${vsday}${vshour}.06h_${ctl}";
    if (! -e "${stageivout}.hdr" || ! -e "${stageivout}.dat") {
      $cmd="$NWPDIR/native/grib_to_native.pl -gribfile $stageivin -leveltype 1 -wgribopt ieee -nativefile $stageivout";
      $cmd="$cmd $verbose_str $debug_str";
      print "At ",scalar localtime()," Running $cmd\n" if $verbose;
      warn ("Error running $cmd\n") if (system($cmd));
    }
  }

  if ($donative) {
    chdir ("$tmpdir/$inittime") or die ("Could not chdir $tmpdir/$inittime\n");
    foreach my $domain ( @domains ) {
      my $sdomain="d".sprintf("%02d",$domain);
      my $wrfin="PRESS-GRIB/d${domain}_${inittime}/d${domain}_${inittime}_${shour}${smin}.grib";
      open(TMPLIST,">d${domain}_${inittime}.list") or die ("Cannot open >d${domain}_${inittime}.list\n");
      open(TMPNARR,">d${domain}_${inittime}_narr.list") or die ("Cannot open >d${domain}_${inittime}_narr.list\n");
      #Loop over types of native files:
      my $reuse=" ";
      for (my $indi=0;$indi < @typ_ind;$indi++) {
	my $ctl=$types[$typ_ind[$indi]];
	my $rotflag=$typ_2d3d[$typ_ind[$indi]];
	my $wrfout="d${domain}_${inittime}_${shour}${smin}_${ctl}";
	$cmd="$NWPDIR/native/grib_to_native.pl -gribfile $wrfin -ctlnam ../${ctl}.in -wgribopt ieee -native $wrfout";
	$cmd="$cmd $verbose_str $debug_str $reuse";
	print "At ",scalar localtime()," Running $cmd\n" if $verbose;
	warn ("Error running $cmd\n") if (system($cmd));

	print TMPLIST "${wrfout}.hdr\n";
	print TMPLIST "${wrfout}.dat\n";

	$reuse="-reuse_gribi";

	#link native narr files to filenames similar to wrf native files
	foreach my $ext ('ctl','dat','hdr') {
	  symlink ("../narr/narr-a_221_${vsyear}${vsmon}${vsday}_${vshour}${vsmin}_${ctl}.${ext}",
		   "narr_${inittime}_${shour}${smin}_${ctl}.${ext}") 
	    unless (-e "narr_${inittime}_${shour}${smin}_${ctl}.${ext}");
	}

	#for GrADS, need to also link original filename into current directory:
	symlink ("../narr/narr-a_221_${vsyear}${vsmon}${vsday}_${vshour}${vsmin}_${ctl}.dat",
		   "narr-a_221_${vsyear}${vsmon}${vsday}_${vshour}${vsmin}_${ctl}.dat") 
	    unless (-e "narr-a_221_${vsyear}${vsmon}${vsday}_${vshour}${vsmin}_${ctl}.dat");

	if ($ctl eq 'pcp_1_all_norot' &&  ($vhour % 6) == 0) {
	  #link native stageiv files to filenames similar to wrf native files
	  foreach my $ext ('ctl','dat','hdr') {
	    symlink ("../stageiv/ST4.${vsyear}${vsmon}${vsday}${vshour}.06h_${ctl}.${ext}",
		     "stageiv_${inittime}_${shour}${smin}.06h_${ctl}.${ext}") 
	      unless (-e "stageiv_${inittime}_${shour}${smin}.06h_${ctl}.${ext}");
	  }

	  #for GrADS, need to also link original filename into current directory:
	  symlink ("../stageiv/ST4.${vsyear}${vsmon}${vsday}${vshour}.06h_${ctl}.dat",
		   "ST4.${vsyear}${vsmon}${vsday}${vshour}.06h_${ctl}.dat") 
	    unless (-e "ST4.${vsyear}${vsmon}${vsday}${vshour}.06h_${ctl}.dat");
	}

	my $wrfnarr="${wrfout}_narr_${rotflag}";
	$cmd="$NWPDIR/native_util/native_interp.x -I${wrfout}.hdr -Gnarr_${inittime}_${shour}${smin}_${ctl}.hdr";
	$cmd="$cmd -O${wrfnarr}.hdr -CT -ST"; #add missing value handling
	$cmd="$cmd -V" if $verbose;
        $cmd="$cmd > ${wrfnarr}.out 2>&1";
	print "At ",scalar localtime()," Running $cmd\n" if $verbose;
	warn ("Error running $cmd\n") if (system($cmd));

	print TMPNARR "${wrfnarr}.hdr\n";
	print TMPNARR "${wrfnarr}.dat\n";

      }
      close(TMPLIST) or warn("Cannot close >d${domain}_${inittime}.list\n");
      $cmd="$NWPDIR/native/native_rotate_winds.pl -listfile d${domain}_${inittime}.list";      
      $cmd="$cmd $verbose_str $debug_str";
      print "At ",scalar localtime()," Running $cmd\n" if $verbose;
      warn ("Error running $cmd\n") if (system($cmd));
      close(TMPNARR) or warn("Cannot close >d${domain}_${inittime}_narr.list\n");
      $cmd="$NWPDIR/native/native_rotate_winds.pl -listfile d${domain}_${inittime}_narr.list";      
      $cmd="$cmd $verbose_str $debug_str";
      print "At ",scalar localtime()," Running $cmd\n" if $verbose;
      warn ("Error running $cmd\n") if (system($cmd));
    }
  }
  $itime++;
  $leadsec = $itime*$interval;
}

1;

sub help{
  my %options=%{$_[0]};
  my %helpopt=%{$_[1]};
  print "
Help for $0 
 This script produces native files from narr and/or wrfout files, using:
 - run_wrfpost.pl to postprocess WRF netcdf eta-level files
 - grib_to_native.pl for converting narr and/or wrfpost output files to native
 - native_interp and native_rotate to convert WRF native to narr-compatible files

 The \"\@types\" option(s), if supplied, must conform to the following convention:
  var_ltype_lvalue_rot, where
  #   var = variable indicator (\"all\" or \"uv\" or \"pcp\")
  #   ltype = level type code (numeric)
  #   lvalue = level value (numeric, or \"all\")
  #   rot = wind rotation indicator (for wrf files) (\"3d\",\"2d\", or \"norot\")

 Options (REQ denotes required arguments):
";
foreach my $key (keys(%options)) {
  print "  $key : $helpopt{$key}\n";
}
die 
"Usage: $0 [options]\n";
}

sub mkctl{
  # decodes type string: var_ltype_lvalue_rot
  #   var = variable indicator ("all" or "uv" )
  #   ltype = level type code (numeric)
  #   lvalue = level value (numeric, or "all")
  #   rot = wind rotation indicator ("3d","2d", or "norot")
  my $types=$_[0];
  my $typ_2d3d=$_[1];
  my $typ_ind=$_[2];
  my $ind3d=-1;
  my $ind2d=-1;
  for (my $i=0;$i<@types;$i++) {
    my @decoded=split(/_/,$$types[$i]);
    open(TMPOUT,">$$types[$i].in") or die("mkctl: Could not open $$types[$i].in\n");
    print TMPOUT "0 \n";			#grid type
    print TMPOUT "$decoded[1] \n";		#level type
    print TMPOUT "-1 -1 -1 -1 -1 -1 -1 -1 -1 \n";	#itimc
    my $lvalue=$decoded[2];
    if ($lvalue eq "all") {
      print TMPOUT "0 \n" ;			#nlevin
    } else {
      print TMPOUT "1 \n" ;			#nlevin
      print TMPOUT "$lvalue \n" ;		#level value
    }
    my $varval=$decoded[0];
  SWITCH_var: {
      if ($varval =~ /^uv$/) {
       print TMPOUT "2 \n" ;			#nparin
       print TMPOUT "33 \n" ;			#var code
       print TMPOUT "34 \n" ;			#var code
       last SWITCH_var;
      }
      if ($varval =~ /^pcp$/) {
       print TMPOUT "2 \n" ;			#nparin
       print TMPOUT "61 \n" ;			#var code: conv pcp
       print TMPOUT "63 \n" ;			#var code: total pcp
       last SWITCH_var;
      }
      #default (including "all"):
       print TMPOUT "0 \n" ;			#nparin
    }
    print TMPOUT "T \n" ;			#lwrite
    close(TMPOUT) or warn("mkctl: Could not close $types[$i].in\n");
    $$typ_2d3d[$i]=$decoded[3];
    if ($$typ_2d3d[$i] eq '3d') {
      die ("mkctl: more than one 3d specified\n") if ($ind3d > -1);
      $ind3d=$i;
    }
    if ($$typ_2d3d[$i] eq '2d') {
      die ("mkctl: more than one 2d specified\n") if ($ind2d > -1);
      $ind2d=$i;
    }
  }
  # reorder types to have 3d/2d at beginning:
  my $i=0;
  if ($ind3d > -1 || $ind2d > -1) {
    die "mkctl: Only one of 3d/2d specified\n" if ($ind3d == -1 || $ind2d == -1);
    $$typ_ind[0]=$ind3d;
    $$typ_ind[1]=$ind2d;
    $i=2;
  }
  for (my $indi=0;$indi<@$types;$indi++) {
    next if ($$typ_2d3d[$indi] eq '2d' || $$typ_2d3d[$indi] eq '3d');
    $$typ_ind[$i]=$indi;
    $i++;
  }
}
